/**
 * Copyright (c) 2014-present, The osquery authors
 *
 * This source code is licensed as defined by the LICENSE file found in the
 * root directory of this source tree.
 *
 * SPDX-License-Identifier: (Apache-2.0 OR GPL-2.0-only)
 */

#include <osquery/core/system.h>

namespace osquery {

void platformSetup() {
  /* No platform-specific logic is needed on POSIX. */
}

void platformTeardown() {
  /* No platform-specific logic is needed on POSIX. */
}
} // namespace osquery
