/**
 * Copyright (c) 2014-present, The osquery authors
 *
 * This source code is licensed as defined by the LICENSE file found in the
 * root directory of this source tree.
 *
 * SPDX-License-Identifier: (Apache-2.0 OR GPL-2.0-only)
 */

#pragma once

#include <string>

#include <osquery/worker/logging/logger.h>

namespace osquery {
class GLOGLogger final : public Logger {
 public:
  static GLOGLogger& instance();
  void log(int severity, const std::string& message) override;
  void vlog(int severity, const std::string& message) override;
};
} // namespace osquery
