/**
 * Copyright (c) 2014-present, The osquery authors
 *
 * This source code is licensed as defined by the LICENSE file found in the
 * root directory of this source tree.
 *
 * SPDX-License-Identifier: (Apache-2.0 OR GPL-2.0-only)
 */

#pragma once

#include <string>

namespace osquery {
class Logger {
 public:
  virtual ~Logger() = default;
  virtual void log(int severity, const std::string& message) = 0;
  virtual void vlog(int severity, const std::string& message) = 0;
};
} // namespace osquery
